import six
from unittest import TestCase

from dark.mutations import getAPOBECFrequencies, mutateString


class TestGetAPOBECFrequencies(TestCase):
    """
    Tests the getAPOBECFrequencies function
    """
    def testCorrectResult(self):
        """
        The function should return the right bases.
        s:  AGTCTAAAGTCATGACTGGTCCCCTTG
        q:      ..AA..T....G....G...
        """
        result = []
        subjectTitle = '333031'
        subjectSequence = 'AGTCTAAAGTCATGACTGGTCCCCTTG'
        subject = '%s \t \t \t %s' % (subjectTitle, subjectSequence)
        result.append(subject)

        queryTitle = 'H9L0NJQ01DXGOQ'
        querySequence = '    ..AA..T....G....G...   '
        query = '%s \t %s' % (queryTitle, querySequence)
        result.append(query)

        bases = getAPOBECFrequencies(result, 'C', 'G', 'cPattern')

        self.assertEqual({'ACA': 0, 'ACC': 0, 'ACG': 0, 'ACT': 1, 'CCA': 0,
                          'CCC': 0, 'CCG': 0, 'CCT': 0, 'GCA': 0, 'GCC': 0,
                          'GCG': 0, 'GCT': 0, 'TCA': 0, 'TCC': 1, 'TCG': 0,
                          'TCT': 0}, bases)


class TestMutateString(TestCase):
    """
    Tests the mutateString function.
    """

    def testZeroLength(self):
        """
        The function should raise when given an empty string
        """
        error = 'Empty original string passed.'
        with six.assertRaisesRegex(self, ValueError, error):
            mutateString('', 1)

    def testTooManyMutationsRequested(self):
        """
        The function should raise when asked to perform more mutations than
        there are in the original string.
        """
        error = 'Cannot make 2 mutations in a string of length 1'
        with six.assertRaisesRegex(self, ValueError, error):
            mutateString('x', 2)

    def testDuplicateReplacementLetter(self):
        """
        The function should raise when given a replacement string that contains
        a duplicate letter.
        """
        error = 'Replacement string contains duplicates'
        with six.assertRaisesRegex(self, ValueError, error):
            mutateString('x', 1, 'aa')

    def testReplacementLengthOneAppearsInOriginal(self):
        """
        The function should raise when given a replacement string that contains
        just one letter if that letter also appears in the original.
        """
        error = 'Impossible replacement'
        with six.assertRaisesRegex(self, ValueError, error):
            mutateString('x', 1, 'x')

    def testZeroReplacements(self):
        """
        The function should return the original string when zero mutations
        are requested.
        """
        self.assertEqual('acgt', mutateString('acgt', 0, 'x'))

    def testOneDeterministicReplacement(self):
        """
        The function should return the correct result when one mutation is
        requested and only one mutation is possible.
        """
        self.assertEqual('c', mutateString('a', 1, 'ac'))

    def testFiveDeterminsticReplacements(self):
        """
        The function should return the correct result when five mutations are
        requested and only one outcome is possible.
        """
        self.assertEqual('ccccc', mutateString('aaaaa', 5, 'ac'))

    def testOneReplacement(self):
        """
        The function should return all possible correct results (and only those
        results) when one mutation is requested and two outcomes are possible.

        NOTE: this test is a bit bad because it's non-deterministic.
        """
        possible = set(['ab', 'ba'])
        seen = set()

        for _ in range(100):
            result = mutateString('aa', 1, 'ab')
            self.assertTrue(result in possible)
            seen.add(result)
        self.assertEqual(seen, possible)

    def testTwoReplacements(self):
        """
        The function should return all possible correct results (and only those
        results) when two mutations are requested and four outcomes are
        possible.

        NOTE: this test is a bit bad because it's non-deterministic.
        """
        possible = set(['ab', 'ba', 'ac', 'ca'])
        seen = set()

        for _ in range(100):
            result = mutateString('aa', 1, 'abc')
            self.assertTrue(result in possible,
                            '%s is not in %r' % (result, possible))
            seen.add(result)
        self.assertEqual(seen, possible)



#!/usr/bin/env python

from __future__ import print_function, division

import argparse
from collections import defaultdict

from dark.sam import samfile

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=('Print SAM/BAM file reference sequence and unmapped read '
                 'counts.'))

parser.add_argument(
    'samFile', metavar='FILENAME',
    help='The name of a SAM/BAM alignment file.')

args = parser.parse_args()


def referenceInfo():
    return {
        'readIds': set(),
        'primaryCount': 0,
        'secondaryCount': 0,
        'supplementaryCount': 0,
    }


referenceReads = defaultdict(referenceInfo)
unmappedCount = 0
readIds = set()
mappingCount = 0

with samfile(args.samFile) as fp:
    for read in fp.fetch():
        mappingCount += 1
        readIds.add(read.query_name)
        if read.is_unmapped:
            unmappedCount += 1
        else:
            stats = referenceReads[read.reference_name]
            stats['readIds'].add(read.query_name)
            if read.is_secondary:
                stats['secondaryCount'] += 1
            elif read.is_supplementary:
                stats['supplementaryCount'] += 1
            else:
                stats['primaryCount'] += 1

totalReads = len(readIds)

print('Found a total of %d read%s, with a total of %d mapping%s and '
      '%d unmapped.' %
      (totalReads, '' if totalReads == 1 else 's',
       mappingCount, '' if mappingCount == 1 else 's', unmappedCount))

for referenceId in sorted(referenceReads):
    stats = referenceReads[referenceId]
    readCount = len(stats['readIds'])
    fraction = (readCount / totalReads) if totalReads else 0.0
    print('%s: %d (%.2f%%) reads mapped. '
          'Primary %d, secondary %d, supplementary %d' %
          (referenceId, readCount, fraction * 100.0,
           stats['primaryCount'], stats['secondaryCount'],
           stats['supplementaryCount']))



from functools import total_ordering

from dark.score import HigherIsBetterScore, LowerIsBetterScore


@total_ordering
class _Base(object):
    """
    Holds information about a matching region from a read alignment.

    You should not use this class directly. Use one of its subclasses,
    either HSP or LSP, depending on whether you want numerically higher
    scores to be considered better (HSP) or worse (LSP).

    Below is an example alignment to show the locations of the six
    start/end offsets.  The upper four are offsets into the subject. The
    lower two are offsets into the read. Note that the read has two gaps
    ('-' characters). All offsets are zero-based and follow the Python
    convention that the 'end' positions are not included in the string.

                           readStartInSubject              readEndInSubject
                           |                               |
                           |                               |
                           |   subjectStart    subjectEnd  |
                           |   |               |           |
                           |   |               |           |
    Subject:  .................ACGTAAAGGCTTAGGT.................
    Read:                  ....ACGTA-AGGCTT-GGT............
                               |               |
                               |               |
                               readStart       readEnd

    Note that the above is just one alignment, and that others are possible
    (e.g., with the read extending beyond the end(s) of the subject, or the
    subject also with gaps in it). The point of the example diagram is to show
    what the six variable names will always refer to, not to enumerate all
    possible alignments (the tests in test/blast/test_hsp.py go through
    many different cases). The classes in this file are just designed to hold
    the variables associated with an HSP and to make it easy to compare them.

    @param readStart: The offset in the read where the match begins.
    @param readEnd: The offset in the read where the match ends.
    @param readStartInSubject: The offset in the subject where the match of
        the read starts.
    @param readEndInSubject: The offset in the subject where the match of
        the read ends.
    @param readFrame: The reading frame for the read, a value from
        {-3, -2, -1, 1, 2, 3} where the sign indicates negative or positive
        sense.
    @param subjectStart: The offset in the subject where the match begins.
    @param subjectEnd: The offset in the subject where the match ends.
    @param subjectFrame: The reading frame for the subject, a value from
        {-3, -2, -1, 1, 2, 3} where the sign indicates negative or positive
        sense.
    @param readMatchedSequence: The matched part of the read. Note that
        this may contain gaps (marked with '-').
    @param subjectMatchedSequence: The matched part of the subject. Note that
        this may contain gaps (marked with '-').
    @param identicalCount: The C{int} number of positions at which the subject
        and query were identical.
    @param positiveCount: The C{int} number of positions at which the subject
        and query had a positive score in the scoring matrix used during
        matching (this is probably only different from the C{identicalCount}
        when matching amino acids (i.e., not nucleotides).
    """
    def __init__(self, readStart=None, readEnd=None, readStartInSubject=None,
                 readEndInSubject=None, readFrame=None, subjectStart=None,
                 subjectEnd=None, subjectFrame=None, readMatchedSequence=None,
                 subjectMatchedSequence=None, identicalCount=None,
                 positiveCount=None):
        self.readStart = readStart
        self.readEnd = readEnd
        self.readStartInSubject = readStartInSubject
        self.readEndInSubject = readEndInSubject
        self.readFrame = readFrame
        self.subjectStart = subjectStart
        self.subjectEnd = subjectEnd
        self.subjectFrame = subjectFrame
        self.readMatchedSequence = readMatchedSequence
        self.subjectMatchedSequence = subjectMatchedSequence
        self.identicalCount = identicalCount
        self.positiveCount = positiveCount

    def __lt__(self, other):
        return self.score < other.score

    def __eq__(self, other):
        return self.score == other.score

    def betterThan(self, score):
        """
        Compare this instance's score with another score.

        @param score: A C{float} score.
        @return: A C{bool}, C{True} if this score is the better.
        """
        return self.score.betterThan(score)

    def toDict(self):
        """
        Get information about the HSP/LSP as a dictionary.

        @return: A C{dict} representation of the HSP/LSP.
        """
        return {
            'readStart': self.readStart,
            'readEnd': self.readEnd,
            'readStartInSubject': self.readStartInSubject,
            'readEndInSubject': self.readEndInSubject,
            'readFrame': self.readFrame,
            'subjectStart': self.subjectStart,
            'subjectEnd': self.subjectEnd,
            'subjectFrame': self.subjectFrame,
            'readMatchedSequence': self.readMatchedSequence,
            'subjectMatchedSequence': self.subjectMatchedSequence,
            'identicalCount': self.identicalCount,
            'positiveCount': self.positiveCount,
        }


class HSP(_Base):
    """
    Holds information about a high-scoring pair from a read alignment.
    Comparisons are done as for BLAST or DIAMOND bit scores (higher is better).

    @param score: The numeric score of this HSP.
    """

    def __init__(self, score, **kwargs):
        _Base.__init__(self, **kwargs)
        self.score = HigherIsBetterScore(score)

    def toDict(self):
        """
        Get information about the HSP as a dictionary.

        @return: A C{dict} representation of the HSP.
        """
        result = _Base.toDict(self)
        result['score'] = self.score.score
        return result


class LSP(_Base):
    """
    Holds information about a low-scoring pair from a read alignment.
    Comparisons are done as for BLAST or DIAMOND e-values (smaller is better).

    @param score: The numeric score of this LSP.
    """

    def __init__(self, score, **kwargs):
        _Base.__init__(self, **kwargs)
        self.score = LowerIsBetterScore(score)

    def toDict(self):
        """
        Get information about the LSP as a dictionary.

        @return: A C{dict} representation of the LSP.
        """
        result = _Base.toDict(self)
        result['score'] = self.score.score
        return result



# From https://samtools.github.io/hts-specs/SAMv1.pdf
CINS, CDEL, CMATCH, CEQUAL, CDIFF = 'IDM=X'


def parseBtop(btopString):
    """
    Parse a BTOP string.

    The format is described at https://www.ncbi.nlm.nih.gov/books/NBK279682/

    @param btopString: A C{str} BTOP sequence.
    @raise ValueError: If C{btopString} is not valid BTOP.
    @return: A generator that yields a series of integers and 2-tuples of
        letters, as found in the BTOP string C{btopString}.
    """
    isdigit = str.isdigit
    value = None
    queryLetter = None
    for offset, char in enumerate(btopString):
        if isdigit(char):
            if queryLetter is not None:
                raise ValueError(
                    'BTOP string %r has a query letter %r at offset %d with '
                    'no corresponding subject letter' %
                    (btopString, queryLetter, offset - 1))
            value = int(char) if value is None else value * 10 + int(char)
        else:
            if value is not None:
                yield value
                value = None
                queryLetter = char
            else:
                if queryLetter is None:
                    queryLetter = char
                else:
                    if queryLetter == '-' and char == '-':
                        raise ValueError(
                            'BTOP string %r has two consecutive gaps at '
                            'offset %d' % (btopString, offset - 1))
                    elif queryLetter == char:
                        raise ValueError(
                            'BTOP string %r has two consecutive identical %r '
                            'letters at offset %d' %
                            (btopString, char, offset - 1))
                    yield (queryLetter, char)
                    queryLetter = None

    if value is not None:
        yield value
    elif queryLetter is not None:
        raise ValueError(
            'BTOP string %r has a trailing query letter %r with '
            'no corresponding subject letter' % (btopString, queryLetter))


def countGaps(btopString):
    """
    Count the query and subject gaps in a BTOP string.

    @param btopString: A C{str} BTOP sequence.
    @raise ValueError: If L{parseBtop} finds an error in the BTOP string
        C{btopString}.
    @return: A 2-tuple of C{int}s, with the (query, subject) gaps counts as
        found in C{btopString}.
    """
    queryGaps = subjectGaps = 0
    for countOrMismatch in parseBtop(btopString):
        if isinstance(countOrMismatch, tuple):
            queryChar, subjectChar = countOrMismatch
            queryGaps += int(queryChar == '-')
            subjectGaps += int(subjectChar == '-')

    return (queryGaps, subjectGaps)


def btop2cigar(btopString, concise=False):
    """
    Convert a BTOP string to a CIGAR string.

    @param btopString: A C{str} BTOP sequence.
    @param concise: If C{True}, use 'M' for matches and mismatches instead
        of the more specific 'X' and '='.

    @raise ValueError: If L{parseBtop} finds an error in C{btopString}.
    @return: A C{str} CIGAR string.
    """
    result = []
    thisLength = thisOperation = currentLength = currentOperation = None

    for item in parseBtop(btopString):
        if isinstance(item, int):
            thisLength = item
            thisOperation = CMATCH if concise else CEQUAL
        else:
            thisLength = 1
            query, reference = item
            if query == '-':
                # The query has a gap. That means that in matching the
                # query to the reference a deletion is needed in the
                # reference.
                assert reference != '-'
                thisOperation = CDEL
            elif reference == '-':
                # The reference has a gap. That means that in matching the
                # query to the reference an insertion is needed in the
                # reference.
                thisOperation = CINS
            else:
                # A substitution was needed.
                assert query != reference
                thisOperation = CMATCH if concise else CDIFF

        if thisOperation == currentOperation:
            currentLength += thisLength
        else:
            if currentOperation:
                result.append('%d%s' % (currentLength, currentOperation))
            currentLength, currentOperation = thisLength, thisOperation

    # We reached the end of the BTOP string. If there was an operation
    # underway, emit it.  The 'if' here should only be needed to catch the
    # case where btopString was empty.
    assert currentOperation or btopString == ''
    if currentOperation:
        result.append('%d%s' % (currentLength, currentOperation))

    return ''.join(result)



def dimensionalIterator(dimensions, maxItems=-1):
    """
    Given a list of n positive integers, return a generator that yields
    n-tuples of coordinates to 'fill' the dimensions. This is like an
    odometer in a car, but the dimensions do not each have to be 10.

    For example: dimensionalIterator((2, 3)) will yield in order
    (0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2). See the tests in
    test_dimension.py for many more examples.

    A dimension may also be given as '*', to provide a dimension that is
    never exhausted. For example, dimensionalIterator(('*', 2)) yields the
    infinite series (0, 0), (0, 1), (1, 0), (1, 1), (2, 0), (2, 1), ....

    maxItems can be used to limit the number of tuples yielded.
    """
    nDimensions = len(dimensions)
    if nDimensions == 0 or maxItems == 0:
        return
    if any(map(lambda x: x != '*' and x <= 0, dimensions)):
        raise ValueError
    odometer = [0, ] * nDimensions
    while maxItems != 0:
        yield tuple(odometer)
        maxItems -= 1
        wheel = nDimensions - 1
        while (dimensions[wheel] != '*' and
               odometer[wheel] == dimensions[wheel] - 1 and
               wheel >= 0):
            odometer[wheel] = 0
            wheel -= 1
        if wheel < 0:
            return
        odometer[wheel] += 1



#!/usr/bin/env python

from __future__ import print_function, division

import sys
from pysam import AlignmentFile

from dark.filter import (
    addFASTAFilteringCommandLineOptions, parseFASTAFilteringCommandLineOptions)
from dark.reads import Reads
from dark.sam import samfile, SAMFilter


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=('Given a SAM/BAM file and a set of filtering criteria '
                     'write filtered SAM/BAM to stdout.'))

    parser.add_argument(
        '--quiet', action='store_true', default=False,
        help='If True, do not print the final summary.')

    parser.add_argument(
        '--bam', action='store_const', const='b', default='',
        help='If given, write (gzip compressed) BAM output.')

    parser.add_argument(
        '--checkResultCount', type=int,
        help=('The number of alignments expected in the output. If this '
              'number is not seen, the script exits with status 1 (and an '
              'error message is also printed, unless --quiet was used).'))

    addFASTAFilteringCommandLineOptions(parser)
    SAMFilter.addFilteringOptions(parser)

    args = parser.parse_args()
    reads = parseFASTAFilteringCommandLineOptions(args, Reads())
    samFilter = SAMFilter.parseFilteringOptions(args, reads.filterRead,
                                                storeQueryIds=True)

    # The following 'if' has a False in it to make it always fail. That's
    # because pysam issue 716 (see below) did not fix the problem as I had
    # hoped. Instead it throws an error if you pass a header that has a
    # modified SQ key with reference ids and there's a difference it
    # doesn't like. It's always safe to use the 'else' below, with the
    # slight downside being that its header will mention all sequence ids,
    # even if you only want a lesser number (via --referenceId). I'm
    # leaving the code here because this is how you would do it, and it
    # might be possible to just copy the 'header' dict below and further
    # adjust it to avoid the pysam error.
    if False and samFilter.referenceIds:
        # Make a header that only includes the wanted reference ids (if
        # any).
        print('If you get a segmentation violation, this is (I think) '
              'due to the following pysam issue '
              'https://github.com/pysam-developers/pysam/issues/716 It is '
              'safer to run without using --referenceId.', file=sys.stderr)
        with samfile(args.samfile) as sam:
            header = sam.header.as_dict()
        referenceIds = samFilter.referenceIds
        sequences = [sequence for sequence in header['SQ']
                     if sequence['SN'] in referenceIds]
        referenceIdsFound = set(sequence['SN'] for sequence in sequences)
        notFound = referenceIds - referenceIdsFound
        if notFound:
            if len(notFound) == len(referenceIds):
                raise ValueError(
                    'None of the provided --referenceId ids were found in %s' %
                    args.samfile)
            else:
                # Just warn about the missing references.
                print(
                    'WARNING: reference%s (%s) given with --referenceId not '
                    'found in %s.' % (
                        '' if len(notFound) == 1 else 's',
                        ', '.join(sorted(notFound)), args.samfile),
                    file=sys.stderr)

        header['SQ'] = sequences
        out = AlignmentFile(sys.stdout, mode='w' + args.bam, header=header)
    else:
        with samfile(args.samfile) as sam:
            out = AlignmentFile(sys.stdout, mode='w' + args.bam, template=sam)

    save = out.write
    kept = 0
    for kept, alignment in enumerate(samFilter.alignments(), start=1):
        save(alignment)

    out.close()

    if not args.quiet:
        total = samFilter.alignmentCount
        print('Read %d alignment%s, kept %d (%.2f%%).' %
              (total, '' if total == 1 else 's', kept,
               0.0 if total == 0 else kept / total * 100.0), file=sys.stderr)

    if args.checkResultCount is not None and kept != args.checkResultCount:
        if not args.quiet:
            print('Did not write the expected %d alignment%s (wrote %d).' %
                  (args.checkResultCount,
                   '' if args.checkResultCount == 1 else 's', kept),
                  file=sys.stderr)
        sys.exit(1)



#!/usr/bin/env python

from __future__ import print_function

import sys

from dark.blast import BlastRecords

if len(sys.argv) < 3:
    print('Usage: %s title record-file.(xml|json)...' % sys.argv[0],
          file=sys.stderr)
    sys.exit(1)
else:
    title = sys.argv[1]
    blastRecords = BlastRecords(sys.argv[2:])
    hits = blastRecords.filterHits(whitelist=set([title]),
                                   negativeTitleRegex='.')
    if title in hits.titles:
        print(hits.titles[title])
    else:
        print('Title %r was not hit by any reads.' % title)



from six import assertRaisesRegex
from unittest import TestCase

from dark.btop import countGaps, parseBtop, btop2cigar


class TestParseBtop(TestCase):
    """
    Tests for the parseBtop function.
    """
    def testOneLetter(self):
        """
        An argument with just one letter must result in a ValueError.
        """
        error = ("^BTOP string 'F' has a trailing query letter 'F' with no "
                 "corresponding subject letter$")
        assertRaisesRegex(self, ValueError, error, list, parseBtop('F'))

    def testConsecutiveGaps(self):
        """
        An argument that has two consecutive gaps characters must result in a
        ValueError.
        """
        error = "^BTOP string '36--' has two consecutive gaps at offset 2$"
        assertRaisesRegex(self, ValueError, error, list, parseBtop('36--'))

    def testConsecutiveIdentical(self):
        """
        An argument that has two consecutive identical (non-gap) characters
        must result in a ValueError.
        """
        error = ("^BTOP string '36AA' has two consecutive identical 'A' "
                 "letters at offset 2$")
        assertRaisesRegex(self, ValueError, error, list, parseBtop('36AA'))

    def testEmpty(self):
        """
        An empty argument must result in an empty list.
        """
        self.assertEqual([], list(parseBtop('')))

    def testOneNumberWithTrailingOneLetter(self):
        """
        An argument that is a number with a single letter must result in a
        ValueError.
        """
        error = ("^BTOP string '36F' has a trailing query letter 'F' with no "
                 "corresponding subject letter$")
        assertRaisesRegex(self, ValueError, error, list, parseBtop('36F'))

    def testThreeLetters(self):
        """
        An argument that has three letters must result in a ValueError.
        """
        error = ("^BTOP string 'ABC' has a trailing query letter 'C' with no "
                 "corresponding subject letter$")
        assertRaisesRegex(self, ValueError, error, list, parseBtop('ABC'))

    def testOneLetterThenANumber(self):
        """
        An argument that is a single letter followed by a number must result
        in a ValueError.
        """
        error = ("^BTOP string 'F36' has a query letter 'F' at offset 0 with "
                 "no corresponding subject letter$")
        assertRaisesRegex(self, ValueError, error, list, parseBtop('F36'))

    def testTwoNumbersWithOneLetterBetween(self):
        """
        An argument that is a number, a single letter, and another number must
        result in a ValueError.
        """
        error = ("^BTOP string '36F77' has a query letter 'F' at offset 2 "
                 "with no corresponding subject letter$")
        assertRaisesRegex(self, ValueError, error, list, parseBtop('36F77'))

    def testOneNumber(self):
        """
        An argument that is just one number must give the expected result.
        """
        self.assertEqual([54], list(parseBtop('54')))

    def testOneNumberThatIsZero(self):
        """
        An argument that is just the number zero must give the expected result.
        """
        self.assertEqual([0], list(parseBtop('0')))

    def testOneNumberWithLeadingZeroes(self):
        """
        An argument that is just one number with leading zeroes must give the
        expected result.
        """
        self.assertEqual([54], list(parseBtop('0054')))

    def testOneQuerySubjectPair(self):
        """
        An argument that is a single query/subject letter pair must give the
        expected result.
        """
        self.assertEqual([('A', 'G')], list(parseBtop('AG')))

    def testTwoQuerySubjectPairs(self):
        """
        An argument that has two query/subject letter pairs must give the
        expected result.
        """
        self.assertEqual([('A', 'G'), ('C', 'T')], list(parseBtop('AGCT')))

    def testOneQuerySubjectPairAndANumber(self):
        """
        An argument that is a single query/subject letter pair followed by a
        number must give the expected result.
        """
        self.assertEqual([('A', 'G'), 33], list(parseBtop('AG33')))


class TestCountGaps(TestCase):
    """
    Tests for the countGaps function.
    """
    def testEmpty(self):
        """
        An argument with an empty string must produce the expected result.
        """
        self.assertEqual((0, 0), countGaps(''))

    def testNumberOnly(self):
        """
        An argument with just a number must produce the expected result.
        """
        self.assertEqual((0, 0), countGaps('88'))

    def testLettersButNoGaps(self):
        """
        An argument with just letters must produce the expected result.
        """
        self.assertEqual((0, 0), countGaps('FGAC'))

    def testOneQueryGap(self):
        """
        An argument with just a query gap must produce the expected result.
        """
        self.assertEqual((1, 0), countGaps('-G'))

    def testOneSubjectGap(self):
        """
        An argument with just a subject gap must produce the expected result.
        """
        self.assertEqual((0, 1), countGaps('G-'))

    def testOneQueryAndOneSubjectGap(self):
        """
        An argument with a query and a subject gap must produce the expected
        result.
        """
        self.assertEqual((1, 1), countGaps('G--G'))

    def testMultipleQueryAndSubjectGaps(self):
        """
        An argument with multiple query and a subject gaps must produce the
        expected result.
        """
        self.assertEqual((3, 2), countGaps('-GG-34-T-T39F-'))


class TestBtop2CigarConcise(TestCase):
    """
    Tests for the btop2cigar function when concise is True.
    """
    def testEmpty(self):
        """
        An empty BTOP string must result in an empty CIGAR string.
        """
        self.assertEqual('', btop2cigar('', concise=True))

    def testMixedMatch(self):
        """
        If a BTOP string specifies that all characters match (in the imprecise
        CIGAR sense where M could be identical characters or not), the CIGAR
        string must be all Ms.
        """
        self.assertEqual('7M', btop2cigar('2GC3AT', concise=True))

    def testRefenceInsertion(self):
        """
        If a BTOP string specifies that the query has character but the
        subject (reference) does not, the CIGAR string must indicate an
        insertion to the reference.
        """
        self.assertEqual('1I', btop2cigar('A-', concise=True))

    def testQueryInsertion(self):
        """
        If a BTOP string specifies that the subject (reference) has character
        but the query does not, the CIGAR string must indicate an deletion in
        the reference.
        """
        self.assertEqual('1D', btop2cigar('-A', concise=True))

    def testAll(self):
        """
        If a BTOP string specifies all possible variations, the CIGAR
        string must be correct.
        """
        self.assertEqual('7M2I4M2D5M',
                         btop2cigar('2GC3ATC-G-4-T-A5', concise=True))


class TestBtop2CigarPrecise(TestCase):
    """
    Tests for the btop2cigar function when concise is False.
    """
    def testEmpty(self):
        """
        An empty BTOP string must result in an empty CIGAR string.
        """
        self.assertEqual('', btop2cigar('', concise=False))

    def testMixedMatch(self):
        """
        If a BTOP string specifies that some characters match and some do
        not, the CIGAR string must be specific about the matches / mismatches.
        """
        self.assertEqual('2=1X3=1X', btop2cigar('2GC3AT', concise=False))

    def testRefenceInsertion(self):
        """
        If a BTOP string specifies that the query has character but the
        subject (reference) does not, the CIGAR string must indicate an
        insertion to the reference.
        """
        self.assertEqual('1I', btop2cigar('A-', concise=False))

    def testQueryInsertion(self):
        """
        If a BTOP string specifies that the subject (reference) has character
        but the query does not, the CIGAR string must indicate an deletion in
        the reference.
        """
        self.assertEqual('1D', btop2cigar('-A', concise=False))

    def testAll(self):
        """
        If a BTOP string specifies all possible variations, the CIGAR
        string must be correct.
        """
        self.assertEqual('2=1X3=1X2I4=2D5=',
                         btop2cigar('2GC3ATC-G-4-T-A5', concise=False))



#!/usr/bin/env python

from __future__ import print_function, division

import sys
import argparse
from os.path import join, exists
from os import mkdir, rename
from math import log10

from dark.reads import (addFASTACommandLineOptions,
                        parseFASTACommandLineOptions)

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=('Split sequences in a FASTA file into separate files, named '
                 'either by their sequence id or numerically.'))

parser.add_argument(
    '--outDir', default='.',
    help='The directory to make the files in.')

parser.add_argument(
    '--verbose', default=False, action='store_true',
    help='If given, print sequence ids as they are processed.')

parser.add_argument(
    '--numeric', default=False, action='store_true',
    help='If given, use numeric filenames, like 1.fasta, 2.fasta, etc.')

parser.add_argument(
    '--noLeadingZeroes', default=False, action='store_true',
    help='If given, numeric filenames will not have leading zeroes.')

parser.add_argument(
    '--force', default=False, action='store_true',
    help='If given, overwrite pre-existing files.')

parser.add_argument(
    '--saveAs', choices=('fasta', 'fastq', 'fasta-ss'),
    help=('The output format. The default is to match the input format, '
          'so there is usually no need to specify this option. It can be '
          'used to force conversion from FASTQ to FASTA'))

addFASTACommandLineOptions(parser)
args = parser.parse_args()
reads = parseFASTACommandLineOptions(args)

if not exists(args.outDir):
    mkdir(args.outDir)

saveAs = (
    args.saveAs or
    (args.fasta and 'fasta') or
    (args.fastq and 'fastq') or
    (args.fasta_ss and 'fasta-ss'))

# Note: we may be reading the FASTA input from stdin, so we cannot read it
# more than once (and I don't want to store it all because it may be very
# large). That's why we do a second phase of processing to renumber the
# files we created if --numeric is used (and --noLeadingZeroes is not).

count = 0
for count, read in enumerate(parseFASTACommandLineOptions(args), start=1):
    id_ = read.id.split()[0]
    if args.numeric:
        base = '%d.fasta' % count
    else:
        id_ = read.id.split()[0]
        base = '%s.fasta' % id_
    if args.verbose:
        print('Writing', base)
    filename = join(args.outDir, base)
    if not args.force and exists(filename):
        print('Will not overwrite pre-existing file "%s". Use --force to '
              'make me. Exiting.' % filename, file=sys.stderr)
        sys.exit(1)
    with open(filename, 'w') as fp:
        print(read.toString(saveAs), end='', file=fp)

# Rename numeric filenames to have leading zeroes.
if count and args.numeric and not args.noLeadingZeroes:
    width = int(log10(count)) + 1
    if width > 1:
        if args.verbose:
            print('Renaming to add leading zeroes')
        for i in range(1, count + 1):
            old = str(i)
            new = '%0*d' % (width, i)
            if old == new:
                # We've reached the point where the new names have as many
                # digits as the old, so we can stop.
                break
            else:
                oldFilename = '%s.fasta' % join(args.outDir, old)
                newFilename = '%s.fasta' % join(args.outDir, new)
                if not args.force and exists(newFilename):
                    print('Will not overwrite pre-existing file "%s". Use '
                          '--force to make me. Exiting.' % newFilename,
                          file=sys.stderr)
                    sys.exit(1)
                if args.verbose:
                    print('  ', oldFilename, '->', newFilename)
                rename(oldFilename, newFilename)



#!/usr/bin/env python

from __future__ import print_function

from collections import defaultdict
from math import log10

from dark.aa import NAMES
from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions
from dark.summarize import sequenceCategoryLengths


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description=('Given FASTA on stdin, write a summary of sequence base '
                     'categories to stdout. It is currently not possible to '
                     'specify the categories on the command line.'))

    parser.add_argument(
        '--baseType', default='nucl', choices=('nucl', 'prot'),
        help='The type of the bases in the input.')

    parser.add_argument(
        '--minLength', default=1, type=int,
        help=('If specified, stretches of reads that are less than this '
              'length will not be reported but will be summarized by an '
              'ellipsis.'))

    parser.add_argument(
        '--concise', action='store_true', default=False,
        help='If specified, do not show the individual sequence regions.')

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = parseFASTACommandLineOptions(args)

    if args.baseType == 'nucl':
        categories = {
            'A': 'nucl',
            'C': 'nucl',
            'G': 'nucl',
            'T': 'nucl',
            '-': 'gap',
        }
        default = 'ambiguous'
    else:
        categories = {}
        for name in NAMES:
            categories[name] = 'aa'
        categories['-'] = 'gap'
        default = 'ambiguous'

    categoryWidth = max(
        [len(category) for category in categories.values()] + [len(default)])

    minLength = args.minLength
    concise = args.concise

    for index, read in enumerate(reads, start=1):
        counts = defaultdict(int)
        readLen = len(read)
        width = int(log10(readLen)) + 1
        if not concise:
            summary = []
            append = summary.append
            offset = 1
        for (category, count) in sequenceCategoryLengths(
                read, categories, defaultCategory=default,
                minLength=minLength):
            counts[category] += count
            if not concise:
                append('    %*d %-*s (offset %*d)' %
                       (width, count, categoryWidth, category, width, offset))
                offset += count
        print('%d: %s (length %d)' % (index, read.id, readLen))
        for category in sorted(counts):
            count = counts[category]
            print('  %-*s: %*d (%6.2f%%)' %
                  (categoryWidth, category, width, count,
                   count / readLen * 100.0))
        if not concise:
            print('\n'.join(summary))



#!/usr/bin/env python

from __future__ import print_function

import argparse

from dark.sam import samReferencesToStr

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description='Print SAM/BAM file reference names and lengths.')

parser.add_argument(
    'samfile', metavar='FILENAME',
    help='The name of a SAM/BAM alignment file.')

print(samReferencesToStr(parser.parse_args().samfile))



#!/usr/bin/env python

# See https://samtools.github.io/hts-specs/SAMv1.pdf for the SAM file
# format specification.

from __future__ import print_function, division

import sys
import argparse
from os.path import basename
from tempfile import TemporaryFile
from functools import partial

from dark import __version__ as VERSION
from dark.btop import btop2cigar
from dark.diamond.conversion import diamondTabularFormatToDicts
from dark.reads import DNARead

FIELDS = 'bitscore btop qframe qqual qseq qseqid slen sstart stitle'

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=('Convert DIAMOND tabular format to SAM. The DIAMOND '
                 'invocation *must* include --outfmt 6 %s' % FIELDS))

parser.add_argument(
    '--printFields', default=False, action='store_true',
    help=('Print the field names in the order that they must be given to '
          'diamond --outfmt 6 to produce correct input for this script, '
          'then exit.'))

parser.add_argument(
    '--mappingQuality', type=int, default=255,
    help=('The mapping quality to use for MAPQ (field 5). The default (255) '
          'indicates that mapping quality information is not available.'))

parser.add_argument(
    '--ram', action='store_true', default=False,
    help=('Do not use a temporary file to hold the non-header SAM output. '
          'This will run faster but use more memory since all non-header SAM '
          'output will be stored in RAM and only written out when the full '
          'header can be determined.'))

parser.add_argument(
    '--keepDescriptions', action='store_true', default=False,
    help=('Do not discard text after the first space in query or subject '
          'sequence ids. Note that. Note that this violates the SAM '
          'specification, but since SAM files are TAB-separated there '
          'is probably only a small chance this will cause any problems '
          'downstream.'))

args = parser.parse_args()

if args.printFields:
    print(FIELDS)
    sys.exit(0)

idOnly = not args.keepDescriptions
mappingQuality = args.mappingQuality
ram = args.ram

if 0 > mappingQuality > 255:
    raise ValueError('Mapping quality must be between 0 and 255 (inclusive)')

referenceLengths = {}

if ram:
    nonHeaderLines = []
    emit = nonHeaderLines.append
else:
    tf = TemporaryFile(mode='w+t', encoding='utf-8')
    emit = partial(print, file=tf)

for match in diamondTabularFormatToDicts(sys.stdin, FIELDS.split()):
    qseqid = match['qseqid'].split()[0] if idOnly else match['qseqid']
    stitle = match['stitle'].split()[0] if idOnly else match['stitle']

    referenceLengths[stitle] = match['slen']

    # If the query frame is less than zero, the match was with a reverse
    # complemented translation of the query. Put the reverse compliment
    # into the SAM output, which seems to be standard / accepted practice
    # based on my web searches. See e.g., https://www.biostars.org/p/131891/
    # for what Bowtie2 does and for some comments on this issue for SAM/BAM
    # files in general.
    if match['qframe'] > 0:
        flag = 0
        qseq = match['qseq']
        qqual = match['qqual'] or '*'
    else:
        flag = 16
        qseq = DNARead('id', match['qseq']).reverseComplement().sequence
        qqual = match['qqual'][::-1] if match['qqual'] else '*'

    emit('\t'.join(map(str, [
        # 1. QNAME
        qseqid,
        # 2. FLAG
        flag,
        # 3. RNAME
        stitle,
        # 4. POS
        match['sstart'],
        # 5. MAPQ
        mappingQuality,
        # 6. CIGAR
        btop2cigar(match['btop']),
        # 7. RNEXT
        '*',
        # 8. PNEXT
        0,
        # 9. TLEN
        0,
        # 10. SEQ
        qseq,
        # 11. QUAL
        qqual,
        # 12. Alignment score
        'AS:i:%d' % int(match['bitscore'])])))


progName = basename(sys.argv[0])

# Print SAM headers.
print('\n'.join(
    [
        '@PG\tID:DIAMOND\tPN:DIAMOND',
        '@PG\tID:%s\tPN:%s (version %s)\tCL:%s %s\tPP:DIAMOND' %
        (progName, progName, VERSION, progName, ' '.join(sys.argv[1:])),
        '@CO\t%s is from the dark-matter package '
        '(https://github.com/acorg/dark-matter/)' % progName,
    ] +
    [
        '@SQ\tSN:%s\tLN:%d' % (name, referenceLengths[name])
        for name in sorted(referenceLengths)
    ]))

# Print non-header lines.
if ram:
    print('\n'.join(nonHeaderLines))
else:
    tf.seek(0)
    for line in tf:
        print(line, end='')
    tf.close()



from __future__ import print_function

import six
import bz2
from json import dumps, loads
from operator import itemgetter
from collections import Counter

from Bio.File import as_handle

from dark.hsp import HSP, LSP
from dark.score import HigherIsBetterScore
from dark.alignments import Alignment, ReadAlignments
from dark.diamond.hsp import normalizeHSP


# The keys in the following are DIAMOND format 6 field names. The values
# are one-argument functions that take a string and return an appropriately
# converted field value.
#
# The following fields are taken from the DIAMOND manual v0.9.22 2018-05-11.
# Fields whose name doesn't appear here will be left as strings.
DIAMOND_FIELD_CONVERTER = {
    'bitscore': float,
    'evalue': float,
    'frame': int,
    'gapopen': int,
    'gaps': int,
    'identicalCount': lambda nident: None if nident is None else int(nident),
    'length': int,
    'mismatch': int,
    'nident': int,
    'pident': float,
    'positive': int,
    'positiveCount': lambda pos: None if pos is None else int(pos),
    'ppos': float,
    'qcovhsp': float,
    'qend': int,
    'qframe': int,
    'qlen': int,
    'qstart': int,
    'score': float,
    'send': int,
    'slen': int,
    'sstart': int,
}


def diamondTabularFormatToDicts(filename, fieldNames):
    """
    Read DIAMOND tabular (--outfmt 6) output and convert lines to dictionaries.

    @param filename: Either a C{str} file name or an open file pointer.
    @param fieldNames: A C{list} or C{tuple} of C{str} DIAMOND field names.
        Run 'diamond -help' to see the full list.
    @raise ValueError: If a line of C{filename} does not have the expected
        number of TAB-separated fields (i.e., len(fieldNames)). Or if
        C{fieldNames} is empty or contains duplicates.
    @return: A generator that yields C{dict}s with keys that are the DIAMOND
        field names and values as converted by DIAMOND_FIELD_CONVERTER.
    """
    nFields = len(fieldNames)
    if not nFields:
        raise ValueError('fieldNames cannot be empty.')

    c = Counter(fieldNames)
    if c.most_common(1)[0][1] > 1:
        raise ValueError(
            'fieldNames contains duplicated names: %s.' %
            (', '.join(sorted(x[0] for x in c.most_common() if x[1] > 1))))

    def identity(x):
        return x

    convertFunc = DIAMOND_FIELD_CONVERTER.get

    with as_handle(filename) as fp:
        for count, line in enumerate(fp, start=1):
            result = {}
            line = line[:-1]
            values = line.split('\t')
            if len(values) != nFields:
                raise ValueError(
                    'Line %d of %s had %d field values (expected %d). '
                    'Line was %r.' %
                    (count,
                     (filename if isinstance(filename, six.string_types)
                      else 'input'),
                     len(values), nFields, line))
            for fieldName, value in zip(fieldNames, values):
                value = convertFunc(fieldName, identity)(value)
                result[fieldName] = value
            yield result


class DiamondTabularFormatReader(object):
    """
    Provide a method that yields parsed tabular records from a file. Store and
    make accessible the global DIAMOND parameters.

    Make sure you run DIAMOND with the right output format. You must use:
        --outfmt 6 qtitle stitle bitscore evalue qframe qseq qstart qend sseq
                   sstart send slen btop nident positive

    @param filename: A C{str} filename or an open file pointer, containing
        DIAMOND tabular records.
    """

    def __init__(self, filename):
        self._filename = filename
        self.application = 'DIAMOND'
        self.params = {
            'application': self.application,
            'reference': ('Buchfink et al., Fast and Sensitive Protein '
                          'Alignment using DIAMOND, Nature Methods, 12, 59-60 '
                          '(2015)'),
            'task': 'blastx',  # TODO: Add support for blastp, if needed.
            'version': 'v0.8.23',
        }

    def records(self):
        """
        Parse the DIAMOND output and yield records. This will be used to read
        original DIAMOND output (either from stdin or from a file) to turn the
        DIAMOND results into Python dictionaries that will then be stored in
        our JSON format.

        @return: A generator that produces C{dict}s containing 'alignments' and
            'query' C{str} keys.
        """
        with as_handle(self._filename) as fp:
            previousQtitle = None
            subjectsSeen = {}
            record = {}
            for line in fp:
                line = line[:-1]
                try:
                    (qtitle, stitle, bitscore, evalue, qframe, qseq,
                     qstart, qend, sseq, sstart, send, slen, btop, nident,
                     positive) = line.split('\t')
                except ValueError as e:
                    # We may not be able to find 'nident' and 'positives'
                    # because they were added in version 2.0.3 and will not
                    # be present in any of our JSON output generated before
                    # that. So those values will be None when reading
                    # DIAMOND output without those fields, but that's much
                    # better than no longer being able to read that data.
                    if six.PY2:
                        error = 'need more than 13 values to unpack'
                    else:
                        error = (
                            'not enough values to unpack (expected 15, '
                            'got 13)')
                    if str(e) == error:
                        (qtitle, stitle, bitscore, evalue, qframe,
                         qseq, qstart, qend, sseq, sstart, send, slen,
                         btop) = line.split('\t')
                        nident = positive = None
                    else:
                        raise
                hsp = {
                    'bits': float(bitscore),
                    'btop': btop,
                    'expect': float(evalue),
                    'frame': int(qframe),
                    'identicalCount': None if nident is None else int(nident),
                    'positiveCount': (
                        None if positive is None else int(positive)),
                    'query': qseq,
                    'query_start': int(qstart),
                    'query_end': int(qend),
                    'sbjct': sseq,
                    'sbjct_start': int(sstart),
                    'sbjct_end': int(send),
                }
                if previousQtitle == qtitle:
                    # We have already started accumulating alignments for this
                    # query.
                    if stitle not in subjectsSeen:
                        # We have not seen this subject before, so this is a
                        # new alignment.
                        subjectsSeen.add(stitle)
                        alignment = {
                            'hsps': [hsp],
                            'length': int(slen),
                            'title': stitle,
                        }
                        record['alignments'].append(alignment)
                    else:
                        # We have already seen this subject, so this is another
                        # HSP in an already existing alignment.
                        for alignment in record['alignments']:
                            if alignment['title'] == stitle:
                                alignment['hsps'].append(hsp)
                                break
                else:
                    # All alignments for the previous query id (if any)
                    # have been seen.
                    if previousQtitle is not None:
                        yield record

                    # Start building up the new record.
                    record = {}
                    subjectsSeen = {stitle}
                    alignment = {
                        'hsps': [hsp],
                        'length': int(slen),
                        'title': stitle,
                    }
                    record['alignments'] = [alignment]
                    record['query'] = qtitle

                    previousQtitle = qtitle

            # Yield the last record, if any.
            if record:
                yield record

    def saveAsJSON(self, fp, writeBytes=False):
        """
        Write the records out as JSON. The first JSON object saved contains
        information about the DIAMOND algorithm.

        @param fp: A C{str} file pointer to write to.
        @param writeBytes: If C{True}, the JSON will be written out as bytes
            (not strings). This is required when we are writing to a BZ2 file.
        """
        if writeBytes:
            fp.write(dumps(self.params, sort_keys=True).encode('UTF-8'))
            fp.write(b'\n')
            for record in self.records():
                fp.write(dumps(record, sort_keys=True).encode('UTF-8'))
                fp.write(b'\n')
        else:
            fp.write(six.u(dumps(self.params, sort_keys=True)))
            fp.write(six.u('\n'))
            for record in self.records():
                fp.write(six.u(dumps(record, sort_keys=True)))
                fp.write(six.u('\n'))


class JSONRecordsReader(object):
    """
    Provide a method that yields JSON records from a file. Store, check, and
    make accessible the DIAMOND parameters.

    @param filename: A C{str} filename containing JSON DIAMOND records.
    @param scoreClass: A class to hold and compare scores (see scores.py).
        Default is C{HigherIsBetterScore}, for comparing bit scores. If you
        are using e-values, pass LowerIsBetterScore instead.
    """
    def __init__(self, filename, scoreClass=HigherIsBetterScore):
        self._filename = filename
        self._scoreClass = scoreClass
        if scoreClass is HigherIsBetterScore:
            self._hspClass = HSP
        else:
            self._hspClass = LSP

        self._open(filename)
        self.diamondTask = self.params['task']

    def _open(self, filename):
        """
        Open the input file. Set self._fp to point to it. Read the first
        line of parameters.

        @param filename: A C{str} filename containing JSON DIAMOND records.
        @raise ValueError: if the first line of the file isn't valid JSON,
            if the input file is empty, or if the JSON does not contain an
            'application' key.
        """
        if filename.endswith('.bz2'):
            if six.PY3:
                self._fp = bz2.open(filename, mode='rt', encoding='UTF-8')
            else:
                self._fp = bz2.BZ2File(filename)
        else:
            self._fp = open(filename)

        line = self._fp.readline()
        if not line:
            raise ValueError('JSON file %r was empty.' % self._filename)

        try:
            self.params = loads(line[:-1])
        except ValueError as e:
            raise ValueError(
                'Could not convert first line of %r to JSON (%s). '
                'Line is %r.' % (self._filename, e, line[:-1]))

    def _dictToAlignments(self, diamondDict, read):
        """
        Take a dict (made by DiamondTabularFormatReader.records)
        and convert it to a list of alignments.

        @param diamondDict: A C{dict}, from records().
        @param read: A C{Read} instance, containing the read that DIAMOND used
            to create this record.
        @return: A C{list} of L{dark.alignment.Alignment} instances.
        """
        alignments = []
        getScore = itemgetter('bits' if self._hspClass is HSP else 'expect')

        for diamondAlignment in diamondDict['alignments']:
            alignment = Alignment(diamondAlignment['length'],
                                  diamondAlignment['title'])
            alignments.append(alignment)
            for diamondHsp in diamondAlignment['hsps']:
                score = getScore(diamondHsp)
                normalized = normalizeHSP(diamondHsp, len(read),
                                          self.diamondTask)
                hsp = self._hspClass(
                    score,
                    readStart=normalized['readStart'],
                    readEnd=normalized['readEnd'],
                    readStartInSubject=normalized['readStartInSubject'],
                    readEndInSubject=normalized['readEndInSubject'],
                    readFrame=diamondHsp['frame'],
                    subjectStart=normalized['subjectStart'],
                    subjectEnd=normalized['subjectEnd'],
                    readMatchedSequence=diamondHsp['query'],
                    subjectMatchedSequence=diamondHsp['sbjct'],
                    # Use blastHsp.get on identicalCount and positiveCount
                    # because they were added in version 2.0.3 and will not
                    # be present in any of our JSON output generated before
                    # that. Those values will be None for those JSON files,
                    # but that's much better than no longer being able to
                    # read all that data.
                    identicalCount=diamondHsp.get('identicalCount'),
                    positiveCount=diamondHsp.get('positiveCount'))

                alignment.addHsp(hsp)

        return alignments

    def readAlignments(self, reads):
        """
        Read lines of JSON from self._filename, convert them to read alignments
        and yield them.

        @param reads: An iterable of L{Read} instances, corresponding to the
            reads that were given to DIAMOND.
        @raise ValueError: If any of the lines in the file cannot be converted
            to JSON.
        @return: A generator that yields C{dark.alignments.ReadAlignments}
            instances.
        """
        if self._fp is None:
            self._open(self._filename)

        reads = iter(reads)

        try:
            for lineNumber, line in enumerate(self._fp, start=2):
                try:
                    record = loads(line[:-1])
                except ValueError as e:
                    raise ValueError(
                        'Could not convert line %d of %r to JSON (%s). '
                        'Line is %r.' %
                        (lineNumber, self._filename, e, line[:-1]))
                else:
                    recordTitle = record['query']
                    while True:
                        # Iterate through the input reads until we find the
                        # one that matches this DIAMOND record.
                        try:
                            read = next(reads)
                        except StopIteration:
                            raise ValueError(
                                'Read generator failed to yield a read '
                                'with id \'%s\' as found in record number %d '
                                'during parsing of DIAMOND output file %r.' %
                                (recordTitle, lineNumber - 1, self._filename))
                        else:
                            # Look for an exact read id / subject title match.
                            # If that doesn't work, allow for the case where
                            # the JSON record has a truncated query (i.e.,
                            # read) id. This covers the situation where a tool
                            # we use (e.g., bwa mem) unconditionally does this
                            # truncation in the output it writes.
                            if (read.id == recordTitle or
                                    read.id.split()[0] == recordTitle):
                                alignments = self._dictToAlignments(record,
                                                                    read)
                                yield ReadAlignments(read, alignments)
                                break
                            else:
                                # This is an input read that had no DIAMOND
                                # matches. So it does not appear in the
                                # DIAMOND's output. Yield an empty
                                # ReadAlignments for it.
                                yield ReadAlignments(read, [])

        finally:
            self._fp.close()
            self._fp = None
